from codecs import open

import numpy as np
from lxml import etree
from lxml.html import Element

ORG_XMI_ID = "{http://www.omg.org/XMI}id"


class AnnotationType(object):
    def __init__(self, value: str, typesystem=None):
        self.value = value
        self.shortname = value.split(".")[-1].upper()
        self._typesystem = typesystem

    def _get_allowed_features(self):
        return self._typesystem.get_features_for_type(self)

    def is_allowed_feature(self, feature: str):
        if self._typesystem is None:
            return True
        else:
            return feature in self._get_allowed_features()


def xml_path_to_annotation_type(xml_path: str):
    return xml_path.replace("{http:///", "").replace("/", ".").replace("}", ".").replace(".ecore", "")


def annotation_type_to_xml_path(annotation_type: str):
    split = annotation_type.split(".")
    return "{http:///%s.ecore}%s" % ("/".join(split[:-1]), split[-1])


class Typesystem(object):
    """
    Reads and models the typesystem.

    Will be necessary to handle Features that are not just single values, but arrays. Not yet in use.
    """

    def __init__(self, typesystem_xml: str = "CorefTypeSystem.xml"):
        with open(typesystem_xml, encoding="utf8") as f:
            typesystem = etree.parse(f)

        all_types = typesystem.findall(".//{http://uima.apache.org/resourceSpecifier}typeDescription")

        self.types = dict()

        for annotation_type in all_types:
            type_name = annotation_type.find("{http://uima.apache.org/resourceSpecifier}name").text
            self.types[type_name] = dict()
            features = annotation_type.find("{http://uima.apache.org/resourceSpecifier}features")
            if features is not None:
                for feature in features:
                    feature_name = feature.find("{http://uima.apache.org/resourceSpecifier}name").text
                    feature_value = feature.find("{http://uima.apache.org/resourceSpecifier}rangeTypeName").text
                    self.types[type_name][feature_name] = feature_value.split(".")[-1]

        print()

    def get_features_for_type(self, annotation_type: AnnotationType):
        assert isinstance(annotation_type, AnnotationType), "Unsupported FeatureType"
        type_value = annotation_type.value
        if type_value not in self.types:
            raise RuntimeError("Type %s not found in Typesystem!" % type_value)

        return self.types[type_value]


class KallimachosAnnotationTypes(object):
    """
    Shortcut to the different types of Annotations.
    Not all types are here yet, just add more if you need them.
    """
    MORPHOLOGY = AnnotationType(
        xml_path_to_annotation_type("{http:///de/uniwue/kalimachos/coref/type.ecore}Morphology"))
    NE = AnnotationType(xml_path_to_annotation_type("{http:///de/uniwue/kalimachos/coref/type.ecore}NamedEntity"))
    SENTENCE = AnnotationType(xml_path_to_annotation_type("{http:///de/uniwue/kalimachos/coref/type.ecore}Sentence"),)
    DS = AnnotationType(xml_path_to_annotation_type("{http:///de/uniwue/kalimachos/coref/type.ecore}DirectSpeech"))
    CHAPTER = AnnotationType(xml_path_to_annotation_type("{http:///de/uniwue/kalimachos/coref/type.ecore}Chapter"))
    PARAGRAPH = AnnotationType(xml_path_to_annotation_type("{http:///de/uniwue/kalimachos/coref/type.ecore}Paragraph"))
    POS = AnnotationType(xml_path_to_annotation_type("{http:///de/uniwue/kalimachos/coref/type.ecore}POS"))
    DEPENDENCY = AnnotationType(
        xml_path_to_annotation_type("{http:///de/uniwue/kalimachos/coref/type.ecore}DependencyParse"))
    VIEW = AnnotationType(xml_path_to_annotation_type("{http:///uima/cas.ecore}View"))
    INVALID_TYPE = AnnotationType("this.type.is.invalid")


class Annotation(object):
    """
    This models an Annotation of any type in the CAS.

    Begin, end and id are saved as attributes for (way) faster access.
    If you add more shortcuts, don't forget to update them in __setitem__.

    All other values can be accessed via annotation[attribute_name] or annotation.get(attribute_name).
    """

    def __init__(self, element: Element, annotation_type: AnnotationType):
        self._element = element
        self.type = annotation_type
        self._data = dict(element.items())
        data = self._data
        self.begin = int(data["begin"])
        self.end = int(data["end"])
        self.id = data[("%s" % ORG_XMI_ID)]

    def __getitem__(self, item):
        return self._data[item if item != "id" else "{http://www.omg.org/XMI}id"]

    def __setitem__(self, key, value):
        internal_key = key if key != "id" else "{http://www.omg.org/XMI}id"

        if not self.type.is_allowed_feature(internal_key):
            raise RuntimeError("Type %s does not have Feature %s!" % (self.type, internal_key))

        self._element.attrib[internal_key] = str(value)
        self._data[key] = str(value)

        if key == "begin":
            self.begin = int(value)
        if key == "end":
            self.end = int(value)
        # if key == "name":
        #    self.Name = value
        if key == "id":
            self.id = value

    def get(self, item):
        return self.__getitem__(item)

    def set(self, key, value):
        self.__setitem__(key, value)

    def __repr__(self):
        return "<Annotation, Name: %s, Data: %s>" % (self._data.get("Name"), self._data.items())

    def __eq__(self, other):
        return self.element == other.element

    @property
    def element(self):
        return self._element


class AnnotationList(object):
    """
    Not currently in use. Probably never will be, as performance would terrible.
    """

    def __init__(self, data: list = list()):
        self._data = data

    def __delitem__(self, key):
        key.getparent().remove(key)
        self._data.remove(key)

    # def __setitem__(self, key, value):
    #    raise RuntimeError("Can not change elements in an AnnotationList! This would break the xmi file")

    def __getitem__(self, item):
        if item in self._data:
            return self._data[item]
        else:
            raise AttributeError(item)


class UnsupportedTypeException(Exception):
    def __init__(self, text):
        self.text = text


class AnnotationIndexDict(object):
    def __init__(self, cas):
        self.cas = cas
        self.data = dict()

    def __getitem__(self, annotation_type: AnnotationType) -> np.array:
        if annotation_type not in self.data:
            ret = np.array(sorted([Annotation(element, annotation_type) for element in
                                   self.cas.tree.findall(annotation_type_to_xml_path(annotation_type.value)) if
                                   int(element.get(ORG_XMI_ID)) in self.cas.members],
                                  key=lambda element: element.begin))
            self.data[annotation_type] = ret
        return self.data[annotation_type]

    def __setitem__(self, key, value):
        self.data.__setitem__(key, value)

    def keys(self):
        return self.data.keys()


class CAS(object):
    """
    This is the CAS class.

    Reads a given xmi file and creates a CAS from it.

    Direct access to the text of the file via cas.sofa_string.

    Provides the AnnotationIndex and a function to write the cas back to a file after modifying it. This is untested!
    """

    def __init__(self, file_path: str, type_system_path: str = "CorefTypeSystem.xml"):
        with open(file_path, encoding="utf8") as f:
            self.tree = etree.parse(f)

        tree = self.tree
        self.sofa = tree.find(".//{http:///uima/cas.ecore}Sofa")
        self.sofa_string = self.sofa.get("sofaString")

        self._members = set(map(int, tree.find(".//{http:///uima/cas.ecore}View").get("members").split(" ")))
        self._max_id = max(self._members)

        # self.all_tags = {tag.tag for tag in all_annotations}

        self.typesystem = Typesystem(type_system_path)

        self._data = AnnotationIndexDict(self)

        self._added_data = []
        self._removed_data = []

    def __hash__(self):
        return hash(self.sofa_string)

    def keys(self):
        return self._data.keys()

    @property
    def data(self) -> AnnotationIndexDict:
        return self._data

    def get_annotation_index(self, annotation_type: AnnotationType):
        """
        Returns a sorted numpy array containing all annotations of the given type.

        :param annotation_type: Which annotation type should be included?
        :return: sorted numpy array containing all annotations of the given type
        """
        if not isinstance(annotation_type, AnnotationType):
            raise UnsupportedTypeException("Unsupported FeatureType")

        return self._data[annotation_type]

    def add_to_members(self, annotation_id: str):
        self._members.add(int(annotation_id))
        if int(annotation_id) > self._max_id:
            self._max_id = annotation_id

    def remove_from_members(self, annotation_id: str):
        self._members.remove(int(annotation_id))
        if self._max_id == int(annotation_id):
            self._max_id = max(self._members)

    def add_annotation(self, annotation: Annotation):
        if not isinstance(annotation, Annotation):
            raise UnsupportedTypeException("Unsupported Annotation")
        annotation_type = annotation.type
        annotations = self.data[annotation_type]
        annotations = np.append(annotations, [annotation])
        self.data[annotation_type] = np.array(sorted(annotations, key=lambda element: element.begin))

        # last_of_type = self.tree.findall(annotation_type.value)[-1]
        # index = self.tree.getroot().index(last_of_type)
        # self.tree.find(annotation_type.value).getparent().insert(index + 1, annotation._element)
        self._added_data.append(annotation)

        self.add_to_members(annotation["id"])

    def remove_annotation(self, annotation: Annotation):
        if not isinstance(annotation, Annotation):
            raise UnsupportedTypeException("Unsupported Annotation")
        annotation_type = annotation.type
        annotations = self.data[annotation_type]
        """:type : np.array"""

        if annotation not in annotations:
            raise RuntimeError("Annotation %s is not in the CAS!" % annotation)
        # noinspection PyTypeChecker
        self.data[annotation_type] = np.delete(annotations, np.argwhere(annotations == annotation))
        self.remove_from_members(annotation["id"])

        # self.tree.getroot().remove(annotation._element)
        self._removed_data.append(annotation)

    def write_to_file(self, file_path: str = "output/test.xmi"):
        """
        Writes the cas back to a file at the given path.

        Untested, could break things!

        :param file_path: path to the file to write as a String
        """
        for annotation in self._added_data:
            annotation_type = annotation.type
            last_of_type = self.tree.findall(annotation_type_to_xml_path(annotation_type.value))[-1]
            index = self.tree.getroot().index(last_of_type)
            self.tree.find(annotation_type_to_xml_path(annotation_type.value)).getparent().insert(index + 1,
                                                                                                  annotation.element)

        for annotation in self._removed_data:
            self.tree.getroot().remove(annotation.element)

        self._added_data, self._removed_data = [], []

        with open(file_path, "w", "utf8") as f:
            self.tree.find(".//{http:///uima/cas.ecore}View").set("members", " ".join(map(str, sorted(self._members))))
            string = etree.tostring(self.tree, encoding="unicode")
            f.write(string)

    @property
    def max_id(self):
        return self._max_id

    @property
    def members(self):
        return self._members
