import re

from cas import CAS, AnnotationType, Annotation, ORG_XMI_ID, KallimachosAnnotationTypes


class CASUtil(object):
    def __init__(self, cas: CAS):
        self._cas = cas
        self._annotation_types = KallimachosAnnotationTypes

    def get_covered_text(self, annotation: Annotation):
        """
        Returns the text covered by the given annotation.

        :param annotation: The annotation covering the text
        :return: The text covered by the given annotation
        """
        begin_index, end_index = annotation.begin, annotation.end
        return self._cas.sofa_string[begin_index:end_index]

    @staticmethod
    def bisect_left(li, insert, lo=0, hi=None):
        """
        Returns the index where the element insert would be inserted into the sorted list li.
        Does not insert anything! Just used to find elements quickly in a sorted list.

        Modified from bisect.bisect_left.

        :param li: the list to insert
        :param insert: the element to find a index for
        :param lo: start index. Unused
        :param hi: end index. Unused
        :return:
        """
        insert_key = insert.begin
        if lo < 0:
            raise ValueError('lo must be non-negative')
        if hi is None:
            hi = len(li)
        while lo < hi:
            mid = (lo + hi) // 2
            if li[mid].begin < insert_key:
                lo = mid + 1
            else:
                hi = mid
        return lo

    # TODO: Not perfectly checked. So far all results were correct, but checking doesn't hurt.
    def get_covered_generator(self, annotation: Annotation, covered_type: AnnotationType):
        """
        Returns a generator for all elements of the given type covered by the given annotation.

        :param annotation: the covering annotation
        :param covered_type: the type of annotations to look for
        :return: a generator for all elements of the given type covered by the given annotation
        """
        end = annotation.end
        annotations = self._cas.get_annotation_index(covered_type)

        index = self.bisect_left(annotations, annotation)

        # if index != len(annotations):
        #    candidate = annotations[index]
        #    if candidate.begin > begin:
        #        index -= 1

        for anno in annotations[index:]:
            if anno.end <= end:
                yield anno
            else:
                return

                # return [annotation for annotation in annotations if
                #        annotation.begin >= begin and annotation.end <= end]

    # TODO: Not perfectly checked. So far all results were correct, but checking doesn't hurt.
    def get_covered(self, annotation: Annotation, covered_type: AnnotationType):
        """
        Returns a list containing all elements of the given type covered by the given annotation.

        :param annotation: the covering annotation
        :param covered_type: the type of annotations to look for
        :return: a list for all elements of the given type covered by the given annotation
        """
        return list(self.get_covered_generator(annotation, covered_type))

    def get_ne_by_id(self, ne_id: str):
        """
        Returns the first named entity annotation with id ne_id.

        Unused.

        :param ne_id:
        :return:
        """
        for ne in self._cas.get_annotation_index(self._annotation_types.NE):
            if ne["ID"] == ne_id:
                return ne

    def get_pos_tag(self, anno: Annotation):
        """
        Returns the pos tag of the given annotation.

        :param anno: the annotation to find a pos tag for
        :return: the pos tag of the given annotation
        """
        return self.get_covered_pos(anno).get("POSTag")

    def get_covered_pos(self, anno: Annotation):
        """
        Returns the POS annotation at the position of the given annotation.

        :param anno: the covering annotation
        :return: POS annotation at the position of the given annotation
        """
        if anno.type is self._annotation_types.POS:
            return anno
        covered = self.get_covered(anno, self._annotation_types.POS)

        if len(covered) == 1:
            return covered[0]
        else:
            return None

    def is_pronoun(self, anno: Annotation) -> bool:
        """
        Returns True if the given annotation covers a pronoun.

        :param anno: the covering annotation
        :return: True iff anno covers a pronoun
        """
        if not anno.type == self._annotation_types.POS:
            covered = self.get_covered(anno, self._annotation_types.POS)
            if len(covered) != 1:
                return False
            anno = covered[0]
        tag = self.get_pos_tag(anno)
        return bool(re.match("PIS|PIAT|PIDAT|PPER|PPOSS|PPOSAT|PRELS|PRELAT|PRF|PWS|PWAT|PWAV|PDS|PDAT", tag))

    # TODO: Not perfectly checked. So far all results were correct, but checking doesn't hurt.
    def get_covering(self, annotation: Annotation, annotation_type: AnnotationType):
        """
        Returns the annotation of the given type covering the given annotation.

        :param annotation: the covered annotation
        :param annotation_type: the type of annotations to look for
        :return: annotation of type type covering the given annotation
        """
        annotation_index = self._cas.get_annotation_index(annotation_type)
        begin = annotation.begin
        end = annotation.end

        index = self.bisect_left(annotation_index, annotation)

        if index == len(annotation_index):
            index -= 1

        candidate = annotation_index[index]
        if candidate.begin <= begin:
            if candidate.end >= end:
                return candidate
            else:
                return None
        else:
            candidate = annotation_index[index - 1]
            if candidate.end >= end:
                return candidate
            else:
                return None

    def get_lemma(self, annotation: Annotation):
        """
        Returns the lemma covered by the given annotation
        :param annotation: the covering annotation
        :return: the lemma covered by the annotation
        """
        pos = self.get_covered_pos(annotation)

        lemma = pos.get("Lemma")
        if lemma == "<unknown>" or lemma is None:
            return self.get_covered_text(pos)
        return lemma

    def create_annotation(self, annotation_type: AnnotationType, data: dict):
        allowed_features = set(self._cas.typesystem.get_features_for_type(annotation_type).keys())
        allowed_features = allowed_features.union({"{http://www.omg.org/XMI}id", "begin", "end", "sofa"})

        unallowed_features = set(data.keys()) - allowed_features
        if not len(unallowed_features) == 0:
            raise RuntimeError(
                "Given features %s contain unallowed features %s!" % (allowed_features, data.keys()))
        if "id" not in data:
            data[ORG_XMI_ID] = str(self._cas.max_id + 1)
        if "sofa" not in data:
            data["sofa"] = "1"

        element = self._cas.tree.getroot().makeelement(_tag=annotation_type.value, attrib=data)

        annotation = Annotation(element, annotation_type)

        return annotation

    @property
    def annotation_types(self):
        return self._annotation_types
